/*   BEM = Batch Equilibrium Model (calculates chemical equilibira)
     Copyright (C) 1996  J. Prenzel (jprenze@gwdg.de)
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>

#include "bem.h"

static int makelt(char *nam, char *chrg);
static int makcpl(char *nam, char *pkeq);
static int makslt(char *wrd[], int nwrds );
static int makvls(char *wrd[], int nwords);
static int makedt(char *typ, char *wrd[], int nwords);
static int makgas(char *wrd[], int nwords);
static int makcec();
static int makpkw(char *pkeq);
static int makxs0(char *typ, char *wrd[], int nwords);
static int makxeq(char *typ, char *wrd[], int nwords);
static int maksdt(char *typ, char *wrd[], int nwords);

/*.................................. make reaction cell "cl" ......*/
int makecl(char clnam[]){
  Cl=malloc(sizeof(CLL));
  if(Cl==NULL){
    bemerr(12,"","","",errno);
    return -32;
  } 
  strncpy(Cl->id,clnam,MAXIDNAM);
  Cl->alf=0.;
  Cl->dalf=1.;
  Cl->step=0;
  Cl->it2cnt=0;
  Cl->it1max = 5000;   /* max. of funct. calls per interstep */  
  Cl->it2max = 250000;  /* max. of funct. calls per step */  
  Cl->relerr = 1.e-6;   /* success criterion in ctrleq */
  Cl->lsierr = 1.e-3;   /* max. error in ln(SI) */
  Cl->itrace = 1;
  Cl->estph = Cl->eststr =-999; 
  return 0;
}

/*................................................ modinit ..........*/ 
/* interpret model-file */ 
int modinit(){
  int i, nwords, lrec, rc, typok;
  char rec[MAXINREC], *wrd[MAXWORDS];
  short l;

  Cl->nelt = Cl->ncpl = Cl->nslt = Cl->niex = Cl->ngas = 0;
  Cl->theta0 = Cl->rho = Cl->sigma = Cl->lnkw = Cl->temp = -999.;


  CALL(rc,makelt("H","1"));
  CALL(rc,makelt("OH","-1"));

  for(i=1;i<=4;i++){ /* 4 passes through model file */
    lrec=0;
    while(1){
      if(fgets(rec,MAXINREC,Mod.fil) == NULL) break;
      typok=0;
      lrec++;
      l=strlen(rec);
      if(*(rec+l-1)=='\n') *(rec+l-1)=0;  /* chop LF */
      wrd[0]=wrd[1]=wrd[2]=wrd[3]=wrd[4]=" ";
      nwords=words(rec,wrd);
      if(nwords==-12){
	sprintf(rec,"%d",lrec);
	bemerr(2,rec,Mod.nam,"",errno);
	return -32;
      }
      if(strcmp(wrd[0],"elt")==0){
	typok=1;
	if(i==1){
	  CALL(rc,makelt(wrd[1],wrd[2]));
	}
      }
      if(strcmp(wrd[0],"cpl")==0){
	typok=1;
	if(i==2){
	  CALL(rc,makcpl(wrd[1],wrd[2]));
	}
      }
      if(strcmp(wrd[0],"pkw")==0){
	typok=1;
	if(i==1){
	  CALL(rc,makpkw(wrd[1]));
	}
      }
      if(strcmp(wrd[0],"slt")==0){
	typok=1;
	if(i==2){
	  CALL(rc,makslt(wrd,nwords));
	}
      }
      if(strcmp(wrd[0],"vls")==0){
	typok=1;
	if(i==4){
	  CALL(rc,makvls(wrd,nwords));   
	}
      }
      if(strcmp(wrd[0],"ctl0")==0){
	typok=1;
	if(i==2){
	  CALL(rc,makedt("ctl0",wrd,nwords));   
	}
      }
      if(strcmp(wrd[0],"estsng")==0){
	typok=1;
	if(i==2){
	  CALL(rc,makedt("estsng",wrd,nwords));   
	}
      }
      if(strcmp(wrd[0],"xpksel")==0){
	typok=1;
	if(i==2){
	  CALL(rc,makxs0("xpksel",wrd,nwords));   
	}
      }
      if(strcmp(wrd[0],"xs0")==0){
	typok=1;
	if(i==2){
	  CALL(rc,makxs0("xs0",wrd,nwords));   
	}
      }
      if(strcmp(wrd[0],"xeql")==0){
	typok=1;
	if(i==2){
	  CALL(rc,makxeq("xeql",wrd,nwords));   
	}
      }
      if(strcmp(wrd[0],"xeqs")==0){
	typok=1;
	if(i==2){
	  CALL(rc,makxeq("xeqs",wrd,nwords));   
	}
      }
      if(strcmp(wrd[0],"cslt0")==0){
	typok=1;
	if(i==3){
	  CALL(rc,maksdt("cslt0",wrd,nwords));   
	}
      }
      if(strcmp(wrd[0],"estslt")==0){
	typok=1;
	if(i==3){
	  CALL(rc,maksdt("estslt",wrd,nwords));   
	}
      }
      if(strcmp(wrd[0],"wptslt")==0){
	typok=1;
	if(i==3){
	  CALL(rc,maksdt("wptslt",wrd,nwords));   
	}
      }
      if(strcmp(wrd[0],"gas")==0){
	typok=1;
	if(i==3){
	  CALL(rc,makgas(wrd,nwords));
	}
      }   
      if(strcmp(wrd[0],"run")==0){
	typok=1;
	if(i==1){
	  CALL(rc,makrun(wrd,nwords));   
	}
      }
      if(*(wrd[0])=='*'){ /* comments */
	typok=1;
      }
      if(nwords==0){ /* empty lines */
	typok=1;
      }
      if(typok==0){
	bemerr(13,wrd[0],wrd[1],wrd[2],errno);
	return -32;
      }
    }
    rewind(Mod.fil);
  }
  return 0;
}
/*............................................ makelt ...............*/
/* allocate memory and enter definitions for new BEM-element
*/
int makelt(char *nam, char *chrg){
  ELT *e;
  char *tail;

  Cl->nelt++;
  if(Cl->nelt>MAXELT){
    bemerr(4,"","","",errno);
    return -32;
  }
  e=malloc(sizeof(ELT));
  if(e==NULL){
    bemerr(3,nam,"","",errno);
    return -32;
  } 
  Cl->el[Cl->nelt-1]=e;
  strncpy((*e).name,nam,MAXELTNAM);
  errno=0;
  tail="";
  (*e).chrg=strtol(chrg,&tail,0);
  if(strlen(tail)>0 || errno !=0) {
    bemerr(8,chrg,nam,"",errno);
    return -32;
  }
if(abs((*e).chrg) > MAXCHRG){
  bemerr(11,chrg,nam,"",errno);
  return -32;
}
  e->ctl0 = e->estsng = -999.;
  e->num = Cl->nelt -1;
  return 0;
}

/*............................................ makcpl ...............*/
/* allocate memory and enter definitions for new BEM-complex
*/
int makcpl(char *nam, char *pkeq  ){
  CPL *c;
  int rc;
  char *tail=NULL,*chrg=NULL;
  double pk;
  Cl->ncpl++;
  if(Cl->ncpl>MAXCPL){
    bemerr(5,"","","",errno);
    return -32;
  }
  c=malloc(sizeof(CPL));
  if(c==NULL){
    bemerr(6,nam,"","",errno);
    return -32;
  } 
  Cl->cp[Cl->ncpl-1]=c;
  strncpy((*c).name,nam,MAXCPLNAM);
  errno=0;
  pk=strtod(pkeq,&tail);
  if(strlen(tail)>0 || errno !=0) {
    bemerr(8,pkeq,nam,"",errno);
    return -32;
  }
  c->lnkds = -pk*M_LN10;
  CALL(rc,cplfrm(nam,(*c).el,(*c).stcf, &(*c).nelt, &(*c).chrg));
  if(abs((*c).chrg) > MAXCHRG){
    sprintf(chrg,"%d",(*c).chrg);
    bemerr(11,chrg,nam,"",errno);
    return -32;
  }
  c->gas=NULL;
  return 0;
}
/*------------------------------------------- makpkw ----------------*/
int makpkw(char *pkeq){ 
  double pk;
  char *tail;
  pk=strtod(pkeq,&tail);
  if(strlen(tail)>0 || errno !=0) {
    bemerr(8,pkeq,"pkw","",errno);
    return -32;
  }
  Cl->lnkw = -pk*M_LN10;
  return 0;
}


/*............................................ makcec ...............*/
/* allocate memory and enter definitions for new BEM-element
*/
int makcec(){
  CEC *x;
  int i;

  Cl->niex++;
  if(Cl->niex>MAXIEX){
    bemerr(26,"","","",0);
    return -32;
  }
  x=malloc(sizeof(CEC));
  if(x==NULL){
    bemerr(25,"cation exchangers","","",0);
    return -32;
  } 
  Cl->xc=x;
  for(i=0;i<MAXELT;i++){
    x->lnksel[i]=-999.;
    x->inparl[i]=-999.;
    x->inpars[i]=-999.;
  }
  x->estxsh=-999.;
  x->xcap=-999.;
  return 0;
}

/*............................................ makslt ...............*/
/* allocate memory and enter definitions for new BEM-salt   
*/
int makslt(char *wrd[], int nwrds ){
  SLT *c;
  int rc,z,i,idok;
  char *tail,chrg[]="      ";
  double pk;
  Cl->nslt++;
  if(Cl->nslt>MAXSLT){
    bemerr(20,"","","",errno);
    return -32;
  }
  c=malloc(sizeof(SLT));
  if(c==NULL){
    bemerr(21 ,wrd[1],"","",errno);
    return -32;
  } 
  Cl->sl[Cl->nslt-1]=c;
  strncpy((*c).name,wrd[1],MAXSLTNAM);

  if(strcmp(wrd[2],"spfun") != 0){
    c->idspf=0;
    errno=0;
    pk=strtod(wrd[2],&tail);
    if(strlen(tail)>0 || errno !=0) {
      bemerr(8,wrd[2],wrd[1],"",errno);
      return -32;
    }
    c->spfpar[0] = pk;
  }
  else{
    idok=0;
    if(strcmp(wrd[3],"1") == 0){
      c->idspf=1;
      idok=1;
    }
    if(strcmp(wrd[3],"2") == 0){
      c->idspf=2;
      idok=1;
    }
    if(idok == 0){
      bemerr(43,wrd[3],"","",0);
      return -32;
    }
    if(nwrds-4 > MAXSPFPAR){
      bemerr(44,wrd[1],"","",0);
      return -32;
    }
    for(i=4;i<nwrds;i++){
      errno=0;
      pk=strtod(wrd[i],&tail);
      if(strlen(tail)>0 || errno !=0) {
	bemerr(8,wrd[2],wrd[1],"",errno);
	return -32;
      }
      c->spfpar[i-4] = pk;
    }
    if(c->idspf == 2){
      setlspl(c->spfpar);
    }
  }

  CALL(rc,cplfrm(wrd[1],(*c).el,(*c).stcf, &(*c).nelt, &z));
  if( z != 0 ){
    sprintf(chrg,"%d",z);
    bemerr(22,chrg,wrd[1],"",errno);
    return -32;
  }
  c->wptfix = c->cslt0  = c->estslt = -999.;
  c->coff = 0.;
  return 0;
}

/*............................................ cplfrm ...............*/
/* cplfrm breaks down a string with a complex or salt definition. It 
returns pointers to the elements contained and stoichiometric coefficients as
arrays starting at given pointers. */

int cplfrm(char *rec, ELT *el[], int ne[], int *npts, int *chrg){
  short npt,k,len,i,ii,s,e,l;
  int j,n,z;
  char elt[MAXELTNAM], c, *tail, num[MAXINTSTR];
  ELT *p=NULL;
  npt=0;
  z=0;  /* charge of comlex */
  len=strlen(rec);

  if(*rec != '(') goto err7;
  k=1;                            /* in () */
  s=1;
  for(i=1;i<=len;i++){
    c=*(rec+i);
    if(k==1){
      if(c==')'){
	k=2;                      /* after () */
	e=i-1;
	n=e-s+1;
	copnb(rec+s,elt,n,MAXELTNAM);
	s=i+1;
/*                                      find element */
	p=NULL;
	for(ii=0;ii<Cl->nelt;ii++){
	  if( strcmp(Cl->el[ii]->name,elt)==0){
	    p=Cl->el[ii];
	    break;
	  }
	}
	if(p==NULL){
	  bemerr(10,elt,rec,"",errno);
	  return -32;
	}
	else{
	  npt++;
	  if(npt>MAXPARTS){
	    bemerr(11,rec,"","",errno);
	    return -32;
	  }
	  *npts=npt;     /* output */
	  el[npt-1]=p;   /* output */
	}
      }
    }
    if(k==2){
      if(c=='(' || c==0){ 
	k=1;
	j=-1;
	e=i-1;
	n=e-s+1;

	copnb(rec+s,num,n,MAXINTSTR);
	if(i>s){
	  errno=0;
	  tail="";
	  j=strtol(num,&tail,0);
	  if(strlen(tail)>0 || errno !=0) {
	    bemerr(9,num,rec,"",errno);
	    return -32;
	  }
 	}
	else{
	  j=1;
	  l=0;
	}
	if(j<1) goto err7;
	s=i+1;
	ne[npt-1]=j;   /* output */
	z+=((*p).chrg)*j;
      }
    }
  }
  *chrg=z;     /* output  */
  return 0;
 err7   : {
   bemerr(7,rec,"","",errno);
   return -32;
 }
}

/*.............................................. getpar ...........
if keyword recognized get numeric (double) parameter from next word  */ 
int getpar(char *key, char *wrd[], int i, int nwords, double *target,
           double factor,int *ier){
  char *tail;
  *ier=0;
  if(strcmp(wrd[i],key)==0){
    if(i>=nwords-1){
      bemerr(14,wrd[i],"","",errno);
      *ier=-32;
      return -32;
    }
    errno=0;
    *target=strtod(wrd[i+1],&tail)*factor;
    if(strlen(tail)>0 || errno !=0) {
      bemerr(8,wrd[i+1],wrd[i],"",errno);
      *ier=-32;
      return -32;
    }
    return 1;
  }
  else{
    return 0;
  }
}

/*.............................................. makvls ............
interprets vls-lines in model file                                 */
int makvls(char *wrd[], int nwords){
  int i,ii,typok,ier;
  int k,rc;
  double pfx;
  i=0;
  while(1){
    i++;
    ii=i;
    if(i>=nwords) break;
    typok=0;

    if(*(wrd[i])=='*')break;

    i+=getpar("estph",wrd,i,nwords,&Cl->estph,1.,&ier);
    if(ier==-32) return -32;
    i+=getpar("estpH",wrd,i,nwords,&Cl->estph,1.,&ier);
    if(ier==-32) return -32;
    i+=getpar("eststr",wrd,i,nwords,&Cl->eststr,1.e-6,&ier);
    if(ier==-32) return -32;
    i+=getpar("theta0",wrd,i,nwords,&Cl->theta0,1.,&ier);
    if(ier==-32) return -32;
    i+=getpar("rho",wrd,i,nwords,&Cl->rho,1.,&ier);
    if(ier==-32) return -32;
    i+=getpar("sigma",wrd,i,nwords,&Cl->sigma0,1.,&ier);
    if(ier==-32) return -32;
    i+=getpar("temp",wrd,i,nwords,&Cl->temp,1.,&ier);
    if(ier==-32) return -32;
    rc=getpar("PCO2",wrd,i,nwords,&pfx,1.e-3,&ier); /* mbar -> bar */
    if(ier==-32) return -32;
    if(rc>0){
      i+=rc;
      for(k=0;k<Cl->ngas;k++){
	if(strcmp(Cl->gs[k]->name,"CO2") == 0){
	  Cl->gs[k]->pfix=pfx;
	  break;
	}
      }
    }
    rc=getpar("estxsh",wrd,i,nwords,&pfx,1.e-3,&ier); 
    if(ier==-32) return -32;
    if(rc>0){
      i+=rc;
      if(Cl->niex>0){
	Cl->xc->estxsh=pfx;
      }
    }

    if(i>ii) typok=1;
    if(typok==0){
      bemerr(15,wrd[i],wrd[0],"",errno);
      return -32;
    }
  }

  return 0;
}
/*.............................................. makrun ............
interprets run-lines in model file                                 */
int makrun(char *wrd[], int nwords){
  int typok;
  int k,rc;
  double a;

  k=1;
  while(k<=nwords-1){
    if(*wrd[k] == '*')break;

    typok=0;

    if(strcmp("itrace",wrd[k]) == 0){
      if(k>=nwords){
	bemerr(14,wrd[k],"","",0);
	return -32;
      }
      k++;
      CALL(rc,dnum(wrd[k],wrd[k-1],&a));
      Cl->itrace=a;
      typok=1;
    }

    if(strcmp("it1max",wrd[k]) == 0){
      if(k>=nwords){
	bemerr(14,wrd[k],"","",0);
	return -32;
      }
      k++;
      CALL(rc,dnum(wrd[k],wrd[k-1],&a));
      Cl->it1max=a;
      typok=1;
    }

    if(strcmp("it2max",wrd[k]) == 0){
      if(k>=nwords){
	bemerr(14,wrd[k],"","",0);
	return -32;
      }
      k++;
      CALL(rc,dnum(wrd[k],wrd[k-1],&a));
      Cl->it2max=a;
      typok=1;
    }

    if(strcmp("relerr",wrd[k]) == 0){
      if(k>=nwords){
	bemerr(14,wrd[k],"","",0);
	return -32;
      }
      k++;
      CALL(rc,dnum(wrd[k],wrd[k-1],&a));
      Cl->relerr=a;
      typok=1;
    }

    if(strcmp("lsierr",wrd[k]) == 0){
      if(k>=nwords){
	bemerr(14,wrd[k],"","",0);
	return -32;
      }
      k++;
      CALL(rc,dnum(wrd[k],wrd[k-1],&a));
      Cl->lsierr=a;
      typok=1;
    }



    if(typok == 0){
      bemerr(15,wrd[k],"run","",0);
      return -32;
    }

    k++;
  }

  return 0;
}

/*.............................................. makgas ............
interprets gas-lines in model file                                 */
int makgas(char *wrd[], int nwords){
  int i,ii,typok,ier,k,kk,rc;
  double pk;
  GAS *g;

  i=0;
  while(1){
    i++;
    ii=i;
    if(i>=nwords) break;
    typok=0;

    if(*(wrd[i])=='*')break;

    rc=getpar("CO2",wrd,i,nwords,&pk,1.,&ier);
    if(ier==-32) return -32;
    if(rc>0){
      i+=rc;
      kk=-1;
      for(k=0;k<Cl->ncpl;k++){
	if(strcmp(Cl->cp[k]->name,"(H)2(CO3)")==0){
	  kk=k;
	  break;
	}
	if(strcmp(Cl->cp[k]->name,"(CO3)(H)2")==0){
	  kk=k;
	  break;
	}
      }
      if(kk==-1){
	bemerr(24,"CO2","(H)2(CO3)","",0);
	return -32;
      }
      g=malloc(sizeof(GAS));
      if(g==NULL){
	bemerr(25 ,"gases","","",0);
	return -32;
      } 
      Cl->ngas++;
      Cl->gs[Cl->ngas-1]=g;
      strncpy(g->name,"CO2",MAXGASNAM);
      g->cpnum=kk;
      g->khen=exp(-pk*M_LN10);
      Cl->cp[kk]->gas = g;
      for(k=0;k<Cl->nelt;k++){
	if(strcmp(Cl->el[k]->name,"CO3") == 0){
	  g->elnum=k;
	  break;
	}
      }
    }

    if(i>ii) typok=1;
    if(typok==0){
      bemerr(15,wrd[i],wrd[0],"",errno);
      return -32;
    }
  }

  return 0;
}


/*.............................................. makedt ............
interprets lines in model file that enter data per element             */
int makedt(char *typ, char *wrd[], int nwords){
  int i,ii,iii,typok,ier;

  i=0;
  while(1){
    i++;
    ii=i;
    if(i>=nwords) break;
    typok=0;

    if(*(wrd[i])=='*')break;

    for(iii=0; iii < Cl->nelt; iii++){
      if(strcmp(typ,"ctl0")==0){                 
	i+=getpar(Cl->el[iii]->name,wrd,i,nwords,&Cl->el[iii]->ctl0
                  ,1.e-6,&ier);
      }

      if(strcmp(typ,"estsng")==0){                 
	i+=getpar(Cl->el[iii]->name,wrd,i,nwords,&Cl->el[iii]->estsng
                  ,1.e-6,&ier);
      }


       if(ier==-32) return -32;
    }

    if(i>ii) typok=1;
    if(typok==0){
      bemerr(15,wrd[i],wrd[0],"",errno);
      return -32;
    }
  }

  return 0;
}

/*.............................................. makxs0 ............*/
int makxs0(char *typ, char *wrd[], int nwords){
  int i,ii,iii,typok,ier;
  
  if(Cl->niex == 0){
    makcec();
    Cl->xc->spec = 'd';   /* direct specification */
  }
  if(Cl->xc->spec == 'i'){;
     bemerr(28,"","","",0);
     return -32;
   }

  i=0;
  while(1){
    i++;
    ii=i;
    if(i>=nwords) break;
    typok=0;

    if(*(wrd[i])=='*')break;

    if(strcmp(typ,"xs0")==0){                 
      i+=getpar("cec",wrd,i,nwords,&Cl->xc->xcap,1.e-3,&ier);
    }
    if(ier==-32) return -32;
 
    for(iii=0; iii < Cl->nelt; iii++){

      if(strcmp(typ,"xpksel")==0){
	i+=getpar(Cl->el[iii]->name,wrd,i,nwords,&Cl->xc->inparl[iii]
                  ,1.,&ier);
      }

      if(strcmp(typ,"xs0")==0){                 
	i+=getpar(Cl->el[iii]->name,wrd,i,nwords,&Cl->xc->inpars[iii]
                  ,1.,&ier);
      }

       if(ier==-32) return -32;
    }

    if(i>ii) typok=1;
    if(typok==0){
      bemerr(15,wrd[i],wrd[0],"",errno);
      return -32;
    }
  }

  return 0;
}

/*.............................................. makxeq ............*/
int makxeq(char *typ, char *wrd[], int nwords){
  int i,ii,iii,typok,ier;
  
  if(Cl->niex == 0){
    makcec();
    Cl->xc->spec = 'i';   /* indirect specification */
  }
  if(Cl->xc->spec == 'd'){;
     bemerr(28,"","","",0);
     return -32;
   }

  i=0;
  while(1){
    i++;
    ii=i;
    if(i>=nwords) break;
    typok=0;

    if(*(wrd[i])=='*')break;

    if(strcmp(typ,"xeql")==0){                 
      i+=getpar("istr",wrd,i,nwords,&Cl->xc->inparl[1],1.e-6,&ier);
      i+=getpar("pH",wrd,i,nwords,&Cl->xc->inparl[0],1.,&ier);
    }
    if(ier==-32) return -32;
 
    if(strcmp(typ,"xeqs")==0){                 
      i+=getpar("H",wrd,i,nwords,&Cl->xc->inpars[0],1.e-3,&ier);
    }
    if(ier==-32) return -32;
 
    for(iii=2; iii < Cl->nelt; iii++){

      if(strcmp(typ,"xeql")==0){
	i+=getpar(Cl->el[iii]->name,wrd,i,nwords,&Cl->xc->inparl[iii]
                  ,1.e-6,&ier); /* input in mol/L */
      }

      if(strcmp(typ,"xeqs")==0){                 
	i+=getpar(Cl->el[iii]->name,wrd,i,nwords,&Cl->xc->inpars[iii]
                  ,1.e-3,&ier); /* input in mmolc/kg */
      }

       if(ier==-32) return -32;
    }

    if(i>ii) typok=1;
    if(typok==0){
      bemerr(15,wrd[i],wrd[0],"",errno);
      return -32;
    }
  }
  return 0;
}

/*.............................................. maksdt ............
interprets lines in model file that enter data per salt                */
int maksdt(char *typ, char *wrd[], int nwords){
  int i,ii,iii,typok,ier;

  i=0;
  while(1){
    i++;
    ii=i;
    if(i>=nwords) break;
    typok=0;

    if(*(wrd[i])=='*'){
      break;
    }

    for(iii=0; iii < Cl->nslt; iii++){
      if(strcmp(typ,"cslt0")==0){                 
	i+=getpar(Cl->sl[iii]->name,wrd,i,nwords,&Cl->sl[iii]->cslt0
                  ,1.e-3,&ier);
      }

      if(strcmp(typ,"estslt")==0){                 
	i+=getpar(Cl->sl[iii]->name,wrd,i,nwords,&Cl->sl[iii]->estslt
                  ,1.e-3,&ier);
      }

      if(strcmp(typ,"wptslt")==0){                 
	i+=getpar(Cl->sl[iii]->name,wrd,i,nwords,&Cl->sl[iii]->wptfix
                  ,1.e-3,&ier);
      }
      if(ier==-32) return -32;
    }

    if(i>ii) typok=1;
    if(typok==0){
      bemerr(15,wrd[i],wrd[0],"",errno);
      return -32;
    }
  }

  return 0;
}

/*............................................ init0 .....................
checks initial values and estimates, sets default values */
int init0(){
  int i,ii,k,z;
  double a,b,c,d,lngam[MAXCHRG+1];
  char r[36];

/* ---------------------------------------- vls ----------------- */  
  if(Cl->theta0 <0){
    Cl->theta0=1.;
  }
  if(Cl->rho <0){
    Cl->rho=1.0;
  }
  if(Cl->sigma <0){
    Cl->sigma=1.0;
  }

  if(Cl->temp <0){
    Cl->temp=25;
  }
  Cl->temp += 273.15; /* Celsius -> Kelvin */

  if(Cl->estph <0){
    bemerr(17,"estpH","5.5","",errno);
    Cl->estph = 5.5;
  }

  if(Cl->eststr <0){
    bemerr(17,"eststr","1000 mol/L","",errno);
    Cl->eststr=0.001;
  }

  if(Cl->lnkw <-998){
    Cl->lnkw = -14.0*M_LN10;
  }

/*----------------------------------------- ctot ------------------*/
  for(i=0; i < Cl->nelt; i++){   
    Cl->el[i]->ctot = 0.;
  }

  for(i=0;i<Cl->nelt;i++){
    if(Cl->el[i]->ctl0 > 0.){
      Cl->el[i]->ctot += Cl->el[i]->ctl0 * Cl->theta0; 
    }
  }

  for(i=0;i<Cl->nslt;i++){
    if(Cl->sl[i]->cslt0 > 0.){
      for(ii=0;ii<Cl->sl[i]->nelt;ii++){
	Cl->sl[i]->el[ii]->ctot += Cl->sl[i]->cslt0 * Cl->sl[i]->stcf[ii]
	* Cl->rho;
      }
    }
  }

  for(i=2;i<Cl->nelt;i++){
    Cl->el[i]->ctoti = Cl->el[i]->ctot;
  }

/*----------------------------------------- PCO2 special ----------------*/ 
  for(i=0;i<Cl->ngas;i++){
    if(strcmp(Cl->gs[i]->name,"CO2")==0 && Cl->gs[i]->pfix > 0.){
      k=Cl->gs[i]->elnum;
      Cl->el[k]->ctot=999.;
      if(Cl->el[k]->estsng < -998.){
	Cl->el[k]->estsng = Cl->gs[i]->pfix * Cl->gs[i]->khen *
	  exp(Cl->cp[Cl->gs[i]->cpnum]->lnkds + 2.*Cl->estph*M_LN10);
	break;
      }
    }
  }



/*----------------------------------------- estsng ----------------*/ 
  for(i=2; i < Cl->nelt; i++){   
    if(Cl->el[i]->estsng < 0.){
      if(Cl->el[i]->ctl0 >0){
	a=Cl->el[i]->ctl0;
      }
      else{
	a=0.001;
      }
      sprintf(r,"%g mol/L",a*1.e6);
      bemerr(19,"estsng",Cl->el[i]->name,r,errno);
      Cl->el[i]->estsng = a;
    }
  }

/*----------------------------------------- estslt ----------------*/ 

  for(i=0; i < Cl->nslt; i++){   
    if(Cl->sl[i]->estslt < 0.){
      if(Cl->sl[i]->cslt0 >0){
	a=Cl->sl[i]->cslt0;
      }
      else{
	a=0.0  ;
      }
      sprintf(r,"%g mmol/kg",a*1.e3);
      bemerr(19,"estslt",Cl->sl[i]->name,r,errno);
      Cl->sl[i]->estslt = a;
    }
  }


/*------------------------------------------ cation exchange -----------*/
  if(Cl->niex==1){
 	
    if(Cl->xc->spec == 'd'){ /*................. direct method .............*/
      if(Cl->xc->xcap <-998.){ 
	Cl->xc->xcap=0.;
      }
      if(Cl->xc->xcap > 0.){
	
	for(i=0; i< Cl->nelt; i++){
	  if(Cl->el[i]->chrg > 0){
	    if(Cl->xc->inparl[i] > -998.){
	      Cl->xc->lnksel[i] = -Cl->xc->inparl[i]*M_LN10;
	    }
	  }
	}
	
	
	Cl->xc->lnksel[0]=0.;         /* 0 <=> H+ */
	for(i=0; i< Cl->nelt; i++){
	  if(Cl->el[i]->chrg > 0){
	    if(Cl->xc->lnksel[i] < -998.){
	      bemerr(27,Cl->el[i]->name,"","",0);
	      Cl->xc->lnksel[i] = 0.;
	    }
	  }
	}
	
	for(i=2;i<Cl->nelt;i++){
	  z=Cl->el[i]->chrg;
	  a=Cl->xc->inpars[i];
	  if(z>0 && a>0.){
	    Cl->el[i]->ctot += a * Cl->xc->xcap / z * Cl->rho;
	  }
	}
	
      }
      
    }
    else{                                   /* indirect method */
      if(Cl->xc->inparl[1] < 0.){
	bemerr(19,"istr","xeql","eststr",0);
/*	c=Cl->eststr;  changed 14.12.95*/
	d=Cl->eststr;
      }
      else{
	d=Cl->xc->inparl[1];
      }
      c=sqrt(d);
      a=-M_LN10*0.5*(c/(1.0+c) - 0.3*d); /* Davies' equation used here */
      for(i=0;i<=MAXCHRG;i++){
	lngam[i]=a*i*i;
      }

      if(Cl->xc->inpars[0] <=-0.){
	bemerr(16,"xs0/H","","",0);
	return -32;
      }
      if(Cl->xc->inparl[0] > -998.){
	b = -Cl->xc->inparl[0]*M_LN10;
      }
      else{
	bemerr(19,"ph","xeql","estph",0);
	b = -Cl->estph*M_LN10;
      }

      Cl->xc->xcap=0.;
      for(i=0;i<Cl->nelt;i++){
	z=Cl->el[i]->chrg;
	if(z > 0){
	  d=Cl->xc->inpars[i];
	  if(d > 0.) {
	    Cl->xc->xcap += d;
	    Cl->el[i]->ctot += d/z*Cl->rho;
	  }
	}
      }

      Cl->xc->lnksel[0]=0.;
      a=log(Cl->xc->inpars[0]/Cl->xc->xcap);
      for(i=2;i<Cl->nelt;i++){
	z=Cl->el[i]->chrg;
	if(z > 0){
	  c=Cl->xc->inparl[i];
	  d=Cl->xc->inpars[i]/Cl->xc->xcap;
	  if(c <= 0. || d <= 0.) {
	    bemerr(27,Cl->el[i]->name,"","",0);
	    Cl->xc->lnksel[i]=0.;
	  }
	  else{
	    Cl->xc->lnksel[i] = (log(c)+lngam[z])/z - log(d)/z + a - b;
	  }
	}
      }
      Cl->xc->estxsh=Cl->xc->inpars[0] / Cl->xc->xcap;
    }

    if(Cl->xc->estxsh <-998.){
      bemerr(17,"estxsh","0.1 %","",0);
      Cl->xc->estxsh=0.1e-2;
    }
    cextab();
  }

/* ------------------------------------------ ctot check --------------*/
  for(i=2; i < Cl->nelt; i++){
    if(Cl->el[i]->ctot <= 0.){
      bemerr(18,Cl->el[i]->name,"","",errno);
      return -32;
    }
  }
  for(i=2;i<Cl->nelt;i++){
    Cl->el[i]->ctoti = Cl->el[i]->ctot;
  }

  return 0;
}












